package extension

import (
	"context"
	"fmt"
	"strings"
	"testing"

	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

// Использование mock.Extension не подходит из-за возникающих циклических импортов
type testServerExtension struct {
	err  error
	name string
}

func (t testServerExtension) GetDescriptor() *ExtensionDescriptor {
	return &ExtensionDescriptor{
		Extension:   t.name,
		Title:       strings.ToTitle(t.name),
		Description: "test extension",
		Version:     "0.0.0",
	}
}

func (t testServerExtension) Install(_ context.Context, _ *InstallRequest) error     { return t.err }
func (t testServerExtension) Check(_ context.Context, _ *CheckRequest) error         { return t.err }
func (t testServerExtension) Uninstall(_ context.Context, _ *UninstallRequest) error { return t.err }
func (t testServerExtension) Action(_ context.Context, _ *ActionRequest) (*ActionResponse, error) {
	return &ActionResponse{}, t.err
}

func getDummyExtension(name string, wantErr ...string) Extension {
	ext := &testServerExtension{name: name}
	if len(wantErr) > 0 {
		ext.err = errors.WithDetail(errors.New(wantErr[0]), "Ошибка")
	}
	return ext
}

func TestServer_Action(t *testing.T) {

	var tests = []struct {
		name      string
		extension Extension
		in        *ActionRequest
		want      *ActionResponse
		wantErr   string
	}{
		{
			name:      "GRPC",
			extension: getDummyExtension("test-extension"),
			in: &ActionRequest{
				Action:  "grpc:///test-extension/test-action",
				SpaceID: "sp",
				EnvID:   "env",
			},
			want: &ActionResponse{State: ResponseDone},
		},
		{
			name:      "invalid schema",
			extension: getDummyExtension("test-extension"),
			in: &ActionRequest{
				Action:  "some:///space/env/coll",
				SpaceID: "sp",
				EnvID:   "env",
			},
			want:    &ActionResponse{State: ResponseDone},
			wantErr: "invalid action",
		},
		{
			name:      "Deprecated call",
			extension: getDummyExtension("test-extension"),
			in: &ActionRequest{
				Action:    "test-action",
				SpaceID:   "sp",
				EnvID:     "env",
				Extension: "test-extension",
			},
			want: &ActionResponse{State: ResponseDone},
		},
		{
			name:      "unknown extension",
			extension: getDummyExtension("test-extension"),
			in: &ActionRequest{
				Action:  "grpc:///test-extension-2/test-action",
				SpaceID: "sp",
				EnvID:   "env",
			},
			want:    nil,
			wantErr: errors.Wrap(ErrUnknownExtension, "test-extension-2").Error(),
		},
		{
			name:      "Deprecated call, without extension",
			extension: getDummyExtension("test-extension"),
			in: &ActionRequest{
				Action:  "test-action",
				SpaceID: "sp",
				EnvID:   "env",
			},
			want:    nil,
			wantErr: "invalid action",
		},
		{
			name:      "Deprecated call, without action and extension)",
			extension: getDummyExtension("test-extension"),
			in: &ActionRequest{
				SpaceID: "sp",
				EnvID:   "env",
			},
			want:    nil,
			wantErr: "invalid action",
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {

			srv := NewServer(NamedExtensions(tt.extension))
			got, err := srv.Action(context.Background(), ActionRequestToPB(tt.in))
			if tt.wantErr != "" {
				require.Error(t, err)
				assert.EqualErrorf(t, err, tt.wantErr, fmt.Sprintf("Action(%v)", tt.in))
				return
			}
			require.NoError(t, err)
			assert.Equalf(t, tt.want, ActionResponseFromPB(got), "Action(%v)", tt.in)
		})
	}
}
