package middleware

import (
	"context"
	"fmt"

	"git.perx.ru/perxis/perxis-go/id"
	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"git.perx.ru/perxis/perxis-go/pkg/invitations"
	"git.perx.ru/perxis/perxis-go/pkg/options"
	logzap "git.perx.ru/perxis/perxis-go/zap"
	"go.uber.org/zap"
)

type loggingMiddleware struct {
	logger *zap.Logger
	next   invitations.Invitations
}

func LoggingMiddleware(logger *zap.Logger) Middleware {
	return func(next invitations.Invitations) invitations.Invitations {
		return &loggingMiddleware{
			next:   next,
			logger: logger.With(logzap.Component("Invitations")),
		}
	}
}

func (m *loggingMiddleware) Create(ctx context.Context, invitation *invitations.Invitation) (created *invitations.Invitation, err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
		logzap.Event(invitations.EventCreate),
		logzap.Object(id.NewSpaceId(invitation.SpaceID)),
	)

	created, err = m.next.Create(ctx, invitation)
	if err != nil {
		logger.Error("Failed to create invitation", zap.Error(err), logzap.Channels(logzap.Userlog, logzap.Syslog))
		return
	}

	logger.Info(fmt.Sprintf("Created invitation '%s'", created.ID), logzap.Channels(logzap.Userlog))

	return created, err
}

func (m *loggingMiddleware) Get(ctx context.Context, invitationId string) (invitation *invitations.Invitation, err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
	)

	invitation, err = m.next.Get(ctx, invitationId)
	if err != nil {
		logger.Error("Failed to get", zap.Error(err))
		return
	}

	return invitation, err
}

func (m *loggingMiddleware) Accept(ctx context.Context, invitationId, userId string) (err error) {
	invitation, err := m.next.Get(ctx, invitationId)
	if err != nil {
		return errors.Wrap(err, "failed to get invitation")
	}

	logger := m.logger.With(
		logzap.Caller(ctx),
		logzap.Event(invitations.EventAccept),
		logzap.Object(id.NewSpaceId(invitation.SpaceID)),
	)

	err = m.next.Accept(ctx, invitationId, userId)
	if err != nil {
		logger.Error(fmt.Sprintf("Failed to accept invitation '%s'", invitationId), zap.Error(err), logzap.Channels(logzap.Userlog, logzap.Syslog))
		return
	}

	logger.Info(fmt.Sprintf("Invitation '%s' accepted by user '%s'", invitationId, userId), logzap.Channels(logzap.Userlog))

	return err
}

func (m *loggingMiddleware) Find(ctx context.Context, filter *invitations.Filter, opts *options.FindOptions) (invitations []*invitations.Invitation, total int, err error) {
	logger := m.logger.With(
		logzap.Caller(ctx),
	)

	invitations, total, err = m.next.Find(ctx, filter, opts)
	if err != nil {
		logger.Error("Failed to find", zap.Error(err))
		return
	}

	return invitations, total, err
}

func (m *loggingMiddleware) Delete(ctx context.Context, invitationId string) (err error) {
	invitation, err := m.next.Get(ctx, invitationId)
	if err != nil {
		return errors.Wrap(err, "failed to get invitation")
	}

	logger := m.logger.With(
		logzap.Caller(ctx),
		logzap.Event(invitations.EventDelete),
		logzap.Object(id.NewSpaceId(invitation.SpaceID)),
	)

	err = m.next.Delete(ctx, invitationId)
	if err != nil {
		logger.Error(fmt.Sprintf("Failed to delete invitation '%s' ", invitationId), zap.Error(err), logzap.Channels(logzap.Userlog, logzap.Syslog))
		return
	}

	logger.Info(fmt.Sprintf("Invitation '%s' deleted", invitationId), logzap.Channels(logzap.Userlog))

	return err
}
