package setup

import (
	"context"
	"testing"

	mockscollections "git.perx.ru/perxis/perxis-go/pkg/collections/mocks"
	"git.perx.ru/perxis/perxis-go/pkg/content"
	"git.perx.ru/perxis/perxis-go/pkg/environments"
	envmocks "git.perx.ru/perxis/perxis-go/pkg/environments/mocks"
	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"github.com/stretchr/testify/mock"

	"git.perx.ru/perxis/perxis-go/pkg/collections"
	"git.perx.ru/perxis/perxis-go/pkg/schema"
	"git.perx.ru/perxis/perxis-go/pkg/schema/field"
	"github.com/stretchr/testify/assert"
)

// Некорректный тест, OverwriteCollection всегда перезаписывает информацию
func TestCollection_UpdateExistingCollection(t *testing.T) {
	tests := []struct {
		name          string
		old, new      *collections.Collection
		wantUpdate    bool
		wantSetSchema bool
		wantErr       bool
	}{
		{
			name:          "Equal collections",
			old:           &collections.Collection{ID: "coll", SpaceID: "sp", EnvID: "env", Schema: schema.New("a", field.String())},
			new:           &collections.Collection{ID: "coll", SpaceID: "sp", EnvID: "env", Schema: schema.New("a", field.String())},
			wantUpdate:    false,
			wantSetSchema: false,
		},
		{
			name:          "Schema changed",
			old:           &collections.Collection{ID: "coll", SpaceID: "sp", EnvID: "env", Name: "Coll", Schema: schema.New("a", field.String())},
			new:           &collections.Collection{ID: "coll", SpaceID: "sp", EnvID: "env", Name: "Coll", Schema: schema.New("b", field.String())},
			wantUpdate:    true,
			wantSetSchema: true,
		},
		{
			name:          "Collection name changed",
			old:           &collections.Collection{ID: "coll", SpaceID: "sp", EnvID: "env", Name: "Coll1", Schema: schema.New("a", field.String())},
			new:           &collections.Collection{ID: "coll", SpaceID: "sp", EnvID: "env", Name: "Coll2", Schema: schema.New("a", field.String())},
			wantUpdate:    true,
			wantSetSchema: false,
		},
		{
			name:          "Collection View changed",
			old:           &collections.Collection{ID: "coll", SpaceID: "sp", EnvID: "env", View: &collections.View{SpaceID: "sp1", EnvID: "env1", CollectionID: "coll1"}},
			new:           &collections.Collection{ID: "coll", SpaceID: "sp", EnvID: "env", View: &collections.View{SpaceID: "sp2", EnvID: "env2", CollectionID: "coll2"}},
			wantUpdate:    true,
			wantSetSchema: false,
		},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			c := new(Collection)
			UpdateExistingCollection()(c)
			collection, update := c.UpdateFunc(new(Setup), tt.old, tt.new)
			assert.Equal(t, tt.wantUpdate, update)
			assert.Equal(t, tt.wantSetSchema, IsSchemaUpdateRequired(tt.old, collection))
		})
	}
}

func TestCollections_InstallCollections(t *testing.T) {
	tests := []struct {
		name            string
		collections     []*collections.Collection
		collectionsCall func(svc *mockscollections.Collections)
		envsCall        func(svc *envmocks.Environments)
		force           bool
		wantErr         func(t *testing.T, err error)
	}{
		{
			name:            "Nil collections",
			collections:     nil,
			collectionsCall: func(svc *mockscollections.Collections) {},
			wantErr: func(t *testing.T, err error) {
				assert.NoError(t, err)
			},
		},
		{
			name: "Install one collection success",
			collections: []*collections.Collection{{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env",
				Schema: schema.New("name", field.String()).ClearState()}},
			collectionsCall: func(svc *mockscollections.Collections) {
				svc.On("Get", mock.Anything, "sp", "env", "1").Return(nil, errors.New("not found")).Once()
				svc.On("Create", mock.Anything,
					&collections.Collection{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env",
						Schema: schema.New("name", field.String()).ClearState()}).
					Return(&collections.Collection{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env",
						Schema: schema.New("name", field.String()).ClearState()}, nil).Once()
				svc.On("SetSchema", mock.Anything, "sp", "env", "1",
					schema.New("name", field.String()).ClearState()).
					Return(nil).Once()
			},
			envsCall: func(svc *envmocks.Environments) {
				svc.On("Migrate", mock.Anything, "sp", "env").Return(nil).Once()
			},
			wantErr: func(t *testing.T, err error) {
				assert.NoError(t, err)
			},
		},
		{
			name:        "Install one view collection success",
			collections: []*collections.Collection{{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env", View: &collections.View{SpaceID: "sp", EnvID: "env", CollectionID: "3"}}},
			collectionsCall: func(svc *mockscollections.Collections) {
				svc.On("Get", mock.Anything, "sp", "env", "1").Return(nil, errors.New("not found")).Once()
				svc.On("Create", mock.Anything, &collections.Collection{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env", View: &collections.View{SpaceID: "sp", EnvID: "env", CollectionID: "3"}}).Return(&collections.Collection{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env", View: &collections.View{SpaceID: "sp", EnvID: "env", CollectionID: "3"}}, nil).Once()
			},
			envsCall: func(svc *envmocks.Environments) {
				svc.On("Migrate", mock.Anything, "sp", "env").Return(nil).Once()
			},
			wantErr: func(t *testing.T, err error) {
				assert.NoError(t, err)
			},
		},
		{
			name:        "Install one collection fails",
			collections: []*collections.Collection{{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String())}},
			collectionsCall: func(svc *mockscollections.Collections) {
				svc.On("Get", mock.Anything, "sp", "env", "1").Return(nil, errors.New("not found")).Once()
				svc.On("Create", mock.Anything,
					&collections.Collection{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String())}).
					Return(nil, errors.New("some error")).Once()
			},
			wantErr: func(t *testing.T, err error) {
				assert.Error(t, err)
				assert.EqualError(t, err, "failed to install collection: some error")
				assert.Contains(t, errors.GetDetail(err), "Возникла ошибка при настройке коллекции space(1)")
			},
		},
		{
			name:        "Install multiple collections success",
			collections: []*collections.Collection{{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String())}, {ID: "2", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String())}},
			collectionsCall: func(svc *mockscollections.Collections) {
				svc.On("Get", mock.Anything, "sp", "env", "1").Return(nil, errors.New("not found")).Once()
				svc.On("Get", mock.Anything, "sp", "env", "2").Return(nil, errors.New("not found")).Once()
				svc.On("Create", mock.Anything, &collections.Collection{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String())}).Return(&collections.Collection{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String())}, nil).Once()
				svc.On("Create", mock.Anything, &collections.Collection{ID: "2", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String())}).Return(&collections.Collection{ID: "2", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String())}, nil).Once()
				svc.On("SetSchema", mock.Anything, "sp", "env", "1", schema.New("name", field.String())).Return(nil).Once()
				svc.On("SetSchema", mock.Anything, "sp", "env", "2", schema.New("name", field.String())).Return(nil).Once()
			},
			envsCall: func(svc *envmocks.Environments) {
				svc.On("Migrate", mock.Anything, "sp", "env").Return(nil).Twice()
			},
			wantErr: func(t *testing.T, err error) {
				assert.NoError(t, err)
			},
		},
		{
			name:        "Install multiple collections fails",
			collections: []*collections.Collection{{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String())}, {ID: "2", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String())}},
			collectionsCall: func(svc *mockscollections.Collections) {
				svc.On("Get", mock.Anything, "sp", "env", "1").Return(nil, errors.New("not found")).Once()
				svc.On("Create", mock.Anything, &collections.Collection{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String())}).Return(nil, errors.New("some error")).Once()
			},
			wantErr: func(t *testing.T, err error) {
				assert.Error(t, err)
				assert.EqualError(t, err, "failed to install collection: some error")
				assert.Contains(t, errors.GetDetail(err), "Возникла ошибка при настройке коллекции space(1)")
			},
		},
		{
			name:        "Update view collection with the same id to new view collection",
			collections: []*collections.Collection{{ID: "1", SpaceID: "sp", EnvID: "env", View: &collections.View{SpaceID: "sp3", EnvID: environments.DefaultEnvironment, CollectionID: "3"}}},
			collectionsCall: func(svc *mockscollections.Collections) {
				svc.On("Get", mock.Anything, "sp", "env", "1").Return(&collections.Collection{ID: "1", SpaceID: "sp", EnvID: "env", View: &collections.View{SpaceID: "sp2", EnvID: environments.DefaultEnvironment, CollectionID: "2"}}, nil).Once()
				svc.On("Update", mock.Anything, &collections.Collection{ID: "1", SpaceID: "sp", EnvID: "env", View: &collections.View{SpaceID: "sp3", EnvID: environments.DefaultEnvironment, CollectionID: "3"}}).Return(nil).Once()
			},
			envsCall: func(svc *envmocks.Environments) {
				svc.On("Migrate", mock.Anything, "sp", "env").Return(nil).Once()
			},
			wantErr: func(t *testing.T, err error) {
				assert.NoError(t, err)
			},
		},
		{
			name:        "Fail to install collection on migrate",
			collections: []*collections.Collection{{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env", Schema: schema.New("name", field.String()).ClearState()}},
			collectionsCall: func(svc *mockscollections.Collections) {
				svc.On("Get", mock.Anything, "sp", "env", "1").Return(nil, errors.New("not found")).Once()
				svc.On("Create", mock.Anything, &collections.Collection{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env",
					Schema: schema.New("name", field.String()).ClearState()}).
					Return(&collections.Collection{ID: "1", SpaceID: "sp", Name: "space", EnvID: "env",
						Schema: schema.New("name", field.String()).ClearState()}, nil).Once()
				svc.On("SetSchema", mock.Anything, "sp", "env", "1",
					schema.New("name", field.String()).ClearState()).Return(nil).Once()
			},
			envsCall: func(svc *envmocks.Environments) {
				svc.On("Migrate", mock.Anything, "sp", "env").Return(errors.New("migrate error")).Once()
			},
			wantErr: func(t *testing.T, err error) {
				assert.Error(t, err)
				assert.EqualError(t, err, "migrate: migrate error")
				assert.Contains(t, errors.GetDetail(err), "Возникла ошибка при миграции данных")
			},
		},
	}

	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			c := &mockscollections.Collections{}
			e := &envmocks.Environments{}
			if tt.collectionsCall != nil {
				tt.collectionsCall(c)
			}
			if tt.envsCall != nil {
				tt.envsCall(e)
			}

			s := NewSetup(&content.Content{Collections: c, Environments: e}, "sp", "env", nil).WithForce(tt.force)
			s.AddCollections(tt.collections)
			tt.wantErr(t, s.InstallCollections(context.Background()))
		})
	}
}
